/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

/*
  kernel loadable modules (aka Kernel EXtensions) support
  
  elf relocation code is based on demo code by Chris Giese

  TODO: make lookup_kernel_symbol use .hash
  (looking through 1000 symbols in a usual way may take a while, although I never noticed this)
*/



#include <tyros.h>
#include <ntx.h>
#include <msgs.h>
#include <rpc.h>
#include <ntxbase.h>
#include <core.h>
#include <vfs.h>
#include <filesystems/filesystems.h>
#include <filesystems/infofs.h>
#include <filesystems/devfs.h>
#include "exec/elf.h"


kex_registry_t *kex_registry;

typedef struct kex_data {
  void *init;
  void *cleanup;
  void *entry;
  void *name;
  void *memarea;
} kex_data_t;

void init_extensions()
{
  kex_registry = malloc(sizeof(kex_registry_t));
  kex_registry->kex = malloc(sizeof(kex_t *));
  kex_registry->count = 0;
}

int register_running_kex (char *filename, kex_data_t *kexdata)
{
  int newcount;
  int newsize;
  int newid;
  kex_t *newkex;

  //  printf("registering kex: %s\n", filename);
  
  newcount = kex_registry->count + 1; 
  newid = newcount - 1;
  newsize = newcount*(sizeof(kex_t *));

  kex_registry->kex = realloc(kex_registry->kex, newsize);
  
  //  printf("new id: %i\n", newid);

  kex_registry->kex[newid] = newkex = malloc(sizeof(kex_t));
  
  newkex->filename = malloc(strlen(filename));
  
  newkex->name = kexdata->name;
  newkex->memarea = kexdata->memarea;
  newkex->init = (unsigned int)kexdata->entry + (unsigned int)kexdata->init;
  newkex->cleanup = (unsigned int)kexdata->entry + (unsigned int)kexdata->cleanup;
  
  strcpy(newkex->filename, filename);
  
  kex_registry->count = newcount;

  return newid;
}


int scan_kex(void *image, kex_data_t *data)
{
  struct elf32_hdr *header;
  struct elf32_shdr *esh, *symtab, *symstrtab;
  struct elf32_sym *symdata;
  int i, syms_count;
  char *symstrings, *symname;

  header = image;
  
  esh = image + header->e_shoff;

  //  printf("kex contains %i sections\n", header->e_shnum);

  for(i = 0; i < header->e_shnum; i++) {
    if(esh[i].sh_type == SHT_SYMTAB) {
      symtab = &esh[i];
    }
  }
  
  symdata = image + symtab->sh_offset;

  syms_count = symtab->sh_size/sizeof(struct elf32_sym);
  
  //  printf("symtab contains %i symbols\n", syms_count);
  
  symstrtab = &esh[symtab->sh_link];
  symstrings = image + symstrtab->sh_offset;
  
  for(i = 0; i < syms_count; i++) {
    symname = symstrings + symdata[i].st_name;
    
    if(!strcmp(symname, "__kex_init")) {
      data->init = symdata[i].st_value;
    }

    if(!strcmp(symname, "__kex_name")) {
      data->name = symdata[i].st_value;
    }

    if(!strcmp(symname, "__kex_cleanup")) {
      data->cleanup = symdata[i].st_value;
    }
    
    //    printf("symbol: %s, value: %X\n", symname, symdata[i].st_value);
    
  }

  return 0;
}

int unload_kex (char *name)
{
  int num;
  int i;
  kex_t *kex;

  //  printf("unloading extension %s\n", name);
  
  for(i = 0; i < kex_registry->count; i++) {
    if(!strcmp(kex_registry->kex[i]->name, name)) {
    
      kex = kex_registry->kex[i];
      
      void (*cleanup)();  
      cleanup = kex->cleanup;
      cleanup();
      
      free_pmem (kex->memarea);
      
      kex_registry->count--;

      kex_registry->kex[i] = kex_registry->kex[kex_registry->count];
      
      return 0;
      
    }
  }

  return -1;
}

int load_kex (char *filename)
{
  char *kexelf;
  file_t file;
  ntx_stat_t statbuf;
  int res;
  void *entry;
  void (*init_func)();
  void (*cleanup_func)();
  kex_data_t kexdata;
  
  //  printf("Loading kernel extension: %s\n", filename);
  
  file = resolve_path (filename);
  
  if(file.id[0] == NULL) {
    err("file not found\n");
    return -1;
  }
  
  stat_file (file, &statbuf);
  
  //  printf("file size: %i bytes\n", (unsigned int) statbuf.st_size);

  
  kexelf = alloc_pmem(mem_to_pages((unsigned int) statbuf.st_size));
  
  kexdata.memarea = kexelf;
  
  read_file (file, 0, statbuf.st_size, kexelf);
  
  res = load_elf_relocatable (kexelf, &entry);
  
  //  printf("load_elf_relocatable: err: %i, entry: %X\n",
  //	 res, entry);
  
  
  if(res == -1) {
    free_pmem(kexelf);
    return -1;
  }

  scan_kex (kexelf, &kexdata);

  kexdata.entry = entry;
  
  init_func = (unsigned int) entry + (unsigned int) kexdata.init;
  cleanup_func = (unsigned int) entry + (unsigned int) kexdata.cleanup;

  kexdata.name = (unsigned int) kexdata.name + (unsigned int) entry;

  init_func();
  
  register_running_kex (filename, &kexdata);
  
  //  free_pmem (kexelf);
  
  return res;
}


unsigned int lookup_kernel_symbol (char *symname)
{
  struct elf32_shdr *symtab;
  struct elf32_shdr *sym_strtab;
  char *sym_strings;
  int syms_count;
  int i;
  void *image = hal.elf_file;
  struct elf32_hdr *header = image;
  struct elf32_shdr *esh;
  struct elf32_sym *symdata;

  //  printf("looking up kernel symbol: %s\n", symname);


  if(* (int *) header != 0x464C457F) {
    failure("Kernel elf image is corrupted\n");
  }


  //  printf("total sections: %i\n", header->e_shnum);
 
  esh = image + header->e_shoff;

  for(i = 0; i < header->e_shnum; i++) {
    if(esh[i].sh_type == SHT_SYMTAB) {
      //      printf("symtab found\n");
      symtab = &esh[i];
    }
  }
  
  sym_strtab = &esh[symtab->sh_link];
  sym_strings = image + sym_strtab->sh_offset;
  symdata = image + symtab->sh_offset;

  syms_count = symtab->sh_size/sizeof(struct elf32_sym);
  
  //  printf("total kernel symbols: %i\n", syms_count);

  //  for(;;);

  for(i = 0; i < syms_count; i++) {
    //    printf("symbol found: %s, val: %X\n", 
    //    sym_strings + symdata[i].st_name,
    //      symdata[i].st_value);
    if(!strcmp(symname, sym_strings + symdata[i].st_name)) {
      return symdata[i].st_value;
    }
  }

  return -1;
}



/*****************************************************************************
ELF RELOCATION ROUTINE

EXPORTS:
int load_elf_relocatable(unsigned char *image, unsigned *entry);
*****************************************************************************/
#include "reloc.h"

int lookup_external_symbol(char *sym_name, unsigned *adr, unsigned uscore);

typedef struct
{
	uint32_t magic		__PACKED__;
	uint8_t bitness		__PACKED__;
	uint8_t endian		__PACKED__;
	uint8_t elf_ver_1	__PACKED__;
	uint8_t res[9]		__PACKED__;
	uint16_t file_type	__PACKED__;
	uint16_t machine	__PACKED__;
	uint32_t elf_ver_2	__PACKED__;
	uint32_t entry_pt	__PACKED__;
	uint32_t phtab_offset	__PACKED__;
	uint32_t shtab_offset	__PACKED__;
	uint32_t flags		__PACKED__;
	uint16_t file_hdr_size	__PACKED__;
	uint16_t phtab_ent_size	__PACKED__;
	uint16_t num_phtab_ents	__PACKED__;
	uint16_t shtab_ent_size	__PACKED__;
	uint16_t num_sects	__PACKED__;
	uint16_t shstrtab_index	__PACKED__;
} elf_file_t;

typedef struct
{
	uint32_t sect_name	__PACKED__;
	uint32_t type		__PACKED__;
	uint32_t flags		__PACKED__;
	uint32_t virt_adr	__PACKED__;
	uint32_t offset		__PACKED__;
	uint32_t size		__PACKED__;
	uint32_t link		__PACKED__;
	uint32_t info		__PACKED__;
	uint32_t align		__PACKED__;
	uint32_t ent_size	__PACKED__;
} elf_sect_t;

typedef struct
{
	uint32_t adr		__PACKED__;
	uint8_t type		__PACKED__;
	uint32_t symtab_index : 24 __PACKED__;
	uint32_t addend		__PACKED__;
} elf_reloc_t;

typedef struct
{
	uint32_t name		__PACKED__;
	uint32_t value		__PACKED__;
	uint32_t size		__PACKED__;
	unsigned type : 4	__PACKED__;
	unsigned binding : 4	__PACKED__;
	uint8_t zero		__PACKED__;
	uint16_t section	__PACKED__;
} elf_sym_t;
/*****************************************************************************
get value of symbol #i
*****************************************************************************/
static int get_elf_sym(unsigned char *image, unsigned i,
		unsigned *sym_val, unsigned symtab_sect)
{
	elf_file_t *file;
	elf_sect_t *sect;
	elf_sym_t *sym;
	char *sym_name;
	unsigned adr;
	int err;

/* point to symbol table */
	file = (elf_file_t *)image;
	if(symtab_sect >= file->num_sects)
	{
		printf("bad symbol table section number %d (max %u)\n",
			symtab_sect, file->num_sects - 1);
		return -1;
	}
	sect = (elf_sect_t *)(image + file->shtab_offset +
		file->shtab_ent_size * symtab_sect);
/* get symbol */
	if(i >= sect->size)
	{
		printf("offset into symbol table (%u) exceeds symbol "
			"table size (%lu)\n", i, sect->size);
		return -1;
	}
	sym = (elf_sym_t *)(image + sect->offset) + i;
/* external symbol */
	if(sym->section == 0)
	{
/* point to string table for this symbol table */
		sect = (elf_sect_t *)(image + file->shtab_offset +
			file->shtab_ent_size * sect->link);
/* get symbol name */
		sym_name = (char *)image + sect->offset + sym->name;
/* ELF binutils for DJGPP: leading underscore
		err = lookup_external_symbol(sym_name, sym_val, 1); */
/* Linux: no leading underscore */
		err = lookup_external_symbol(sym_name, sym_val, 0);
		if(err != 0)
			return err;
	}
/* internal symbol */
	else
	{
		sect = (elf_sect_t *)(image + file->shtab_offset +
			file->shtab_ent_size * sym->section);
		adr = (unsigned)image + sect->offset;
		*sym_val = sym->value + adr;
	}
	return 0;
}
/*****************************************************************************
*****************************************************************************/
static int do_elf_reloc(unsigned char *image, elf_reloc_t *reloc,
		elf_sect_t *sect)
{
	unsigned t_adr, sym_val;
	elf_sect_t *t_sect;
	elf_file_t *file;
	uint32_t *where;
	int err;

/* get address of target section */
	file = (elf_file_t *)image;
	t_sect = (elf_sect_t *)(image + file->shtab_offset +
		file->shtab_ent_size * sect->info);
	t_adr = (unsigned)image + t_sect->offset;
/* point to relocation */
	where = (uint32_t *)(t_adr + reloc->adr);
/* get symbol */
	err = get_elf_sym(image, reloc->symtab_index, &sym_val, sect->link);
	if(err != 0)
		return err;
	switch(reloc->type)
	{
/* absolute reference
Both ELF.H and objdump call this "R_386_32" */
	case 1: /* S + A */
		*where = sym_val + *where;
		break;
/* EIP-relative reference
Both ELF.H and objdump call this "R_386_PC32" */
	case 2: /* S + A - P */
		*where = sym_val + *where - (unsigned)where;
		break;
	default:
		printf("unknown/unsupported relocation type %u "
			"(must be 1 or 2)\n", reloc->type);
		return -1;
	}
	return 0;
}
/*****************************************************************************
*****************************************************************************/
int load_elf_relocatable(unsigned char *image, unsigned *entry)
{
	unsigned s, r, reloc_size;
	unsigned char *bss;
	elf_reloc_t *reloc;
	elf_sect_t *sect;
	elf_file_t *file;
	int err;


	  /* validate */
	file = (elf_file_t *)image;
	if(file->magic != 0x464C457FL) /* "ELF" */
	{
		printf("File is not relocatable ELF; has bad magic value "
			"0x%lX (should be 0x464C457F)\n", file->magic);
		return +1;
	}
	if(file->bitness != 1)
	{
		printf("File is 64-bit ELF, not 32-bit\n");
		return -1;
	}
	if(file->endian != 1)
	{
		printf("File is big endian ELF, not little\n");
		return -1;
	}
	if(file->elf_ver_1 != 1)
	{
		printf("File has bad ELF version %u\n", file->elf_ver_1);
		return -1;
	}
	if(file->file_type != 1)
	{
		printf("File is not relocatable ELF (could be "
			"executable, DLL, or core file)\n");
		return -1;
	}
	if(file->machine != 3)
	{
		printf("File is not i386 ELF\n");
		return -1;
	}
	if(file->elf_ver_2 != 1)
	{
		printf("File has bad ELF version %lu\n", file->elf_ver_2);
		return -1;
	}
/* find the BSS and allocate memory for it
This must be done BEFORE doing any relocations */
	for(s = 0; s < file->num_sects; s++)
	{
		sect = (elf_sect_t *)(image + file->shtab_offset +
				file->shtab_ent_size * s);
		if(sect->type != 8)	/* NOBITS */
			continue;
		r = sect->size;
		
		if(r != 0) {

		bss = calloc(1, r);
		if(bss == NULL)
		{
			printf("Can't allocate %u bytes for BSS\n",
				r);
			return -1;
		}
		sect->offset = bss - image;

		}
		break;
	}
/* for each section... */
	for(s = 0; s < file->num_sects; s++)
	{
		sect = (elf_sect_t *)(image + file->shtab_offset +
				file->shtab_ent_size * s);
/* is it a relocation section?
xxx - we don't handle the extra addend for RELA relocations */
		if(sect->type == 4)	/* RELA */
			reloc_size = 12;
		else if(sect->type == 9)/* REL */
			reloc_size = 8;
		else
			continue;
/* for each relocation... */
		for(r = 0; r < sect->size / reloc_size; r++)
		{
			reloc = (elf_reloc_t *)(image + sect->offset +
				reloc_size * r);
			err = do_elf_reloc(image, reloc, sect);
			if(err != 0)
				return err;
		}
	}
/* find start of .text and make it the entry point */
	(*entry) = 0;
	for(s = 0; s < file->num_sects; s++)
	{
		sect = (elf_sect_t *)(image + file->shtab_offset +
				file->shtab_ent_size * s);
		if((sect->flags & 0x0004) == 0)
			continue;
		(*entry) = (unsigned)image + sect->offset;
		break;
	}
	if((*entry) == 0)
	{
		printf("Can't find section .text, so entry point is unknown\n");
		return -1;
	}
	return 0;
}
int lookup_external_symbol(char *sym_name, unsigned *adr, unsigned uscore)
{
  *adr = lookup_kernel_symbol (sym_name);
  
  if(*adr != -1) {
    return 0;
  } else {
    return -1;
  }
}
