/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <ntx.h>
#include <vfs.h>

/*
  VFS layer:
  physical location of every file is specified by its 'namespace entry' (NSE)
  Each NSE contains list of sub-NSE (>= 0) AND list of names for these NSEs
  
  file's NSE tells everything about this file EXCEPT its name
  (i.e: filename is ONLY determined by the directory the file situated in)
  That means: file can have ANY number of names in ANY number of directories
  (and these directories can be situated in ANY directory and belong to ANY
  filesystem.
  
  During init the VFS layer builds the following directory structure:

/  -- root directory
  /dev   -- device files
  /mnt   -- mount points
      /rd -- ramdisk mountpoint  
  /sys
      /proc -- proc filesystem
      /info -- info filesystem
      
*/
vfs_registry_t *vfs_registry;

void init_vfs_registry()
{
vfs_registry = calloc(sizeof(vfs_registry_t) * 1, 1);
  
vfs_registry->nhandlers = 1;
vfs_registry->vfs_handler = calloc(sizeof(vfs_handler_t), 1);

report("VFS registry is ready\n");
}


int read_dummyfs(id_t id, fsize_t start, fsize_t size, int buffer, file_t device)
{
  return -1;
}

int write_dummyfs(id_t id, fsize_t start, fsize_t size, int buffer, file_t device)
{
  return -1;
}

int stat_dummyfs(id_t id, ntx_stat_t *buffer, file_t device)
{
  return -1;
}

int unlink_dummyfs(id_t id, file_t device)
{
  return -1;
}

int creat_dummyfs(char *file, int mode, file_t device)
{
  return -1;
}

int resolve_dummyfs(file_t device, char *path, char **newpath)
{
  return -1;
}

int mount_dummyfs(nsentry_t *device, nsentry_t *point)
{
  return -1;
}

int mkdir_dummyfs(id_t id, char *name, int mode, file_t device)
{
  return -1;
}

int get_fs_by_name(char *name)
{
  int i;
  
  for(i = 0; i < vfs_registry->nhandlers; i++) {
    if(vfs_registry->vfs_handler[i] != NULL) {
      if(vfs_registry->vfs_handler[i]->name != NULL) {
	if(!strcmp(vfs_registry->vfs_handler[i]->name, name)) {
	  return i;
	}
      }
    }
  }

  return -1;
}

void add_vfs_handler (int id, vfs_handler_t *vfs_handler)
{
  int newsize, oldsize;
  void *oldh;

  if(id >= vfs_registry->nhandlers) {
    
    newsize = (id + 1)*sizeof(vfs_handler_t *);
    oldsize = vfs_registry->nhandlers*sizeof(vfs_handler_t *);

    oldh = vfs_registry->vfs_handler;
    
    vfs_registry->vfs_handler = calloc(1, newsize);
    memcpy(vfs_registry->vfs_handler, oldh, oldsize);
    free(oldh);
    
    //    vfs_registry->vfs_handler = realloc(vfs_registry->vfs_handler, newsize);
    
    vfs_registry->nhandlers = id + 1;
  }

  
  if(vfs_handler->read_func == NULL) {
    vfs_handler->read_func = &read_dummyfs;
  }
  if(vfs_handler->write_func == NULL) {
    vfs_handler->write_func = &write_dummyfs;
  }

  if(vfs_handler->stat_func == NULL) {
    vfs_handler->stat_func = &stat_dummyfs;
  }
  
  if(vfs_handler->unlink_func == NULL) {
    vfs_handler->unlink_func = &unlink_dummyfs;
  }

  if(vfs_handler->resolve_func == NULL) {
    vfs_handler->resolve_func = &resolve_dummyfs;
  }

  if(vfs_handler->creat_func == NULL) {
    vfs_handler->creat_func = &creat_dummyfs;
  }

  if(vfs_handler->mount_func == NULL) {
    vfs_handler->mount_func = &mount_dummyfs;
  }

  if(vfs_handler->mkdir_func == NULL) {
    vfs_handler->mkdir_func = &mkdir_dummyfs;
  }

  vfs_registry->vfs_handler[id] = vfs_handler;
}
