/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

//Based on Free-DOS 32 code(floppy.c)

#include <tyros.h>
#include <ntx.h>
#include <ntxbase.h>
#include <filesystems.h>
#include <devfs.h>
#include <assert.h>
#include <blkdev.h>

#define CMOS_REG 	0x70
#define CMOS_DAT 	0x71

#define DMA_MEMORY_AREA 0x9000

#define FDC_PRI_IO	0x3F0  /* Base port of the primary controller   */
#define FDC_SEC_IO 	0x370  /* Base port of the secondary controller */
#define FDC_DOR  	0x002  /* RW: Digital Output Register           */
#define FDC_MSR  	0x004  /* R : Main Status Register              */
#define FDC_DRS  	0x004  /* W : Data Rate Select Register         */
#define FDC_DATA 	0x005  /* RW: Data Register                     */
#define FDC_DIR  	0x007  /* R : Digital Input Register            */
#define FDC_CCR  	0x007  /* W : Configuration Control Register    */

#define FDC_IRQ		0x26
#define FDC_IRQ_TIMEOUT 1000000

#define CMD_SPECIFY	0x03   /* Specify drive timings   */
#define CMD_WRITE 	0xC5   /* Write data (+ MT,MFM)   */
#define CMD_READ   	0xE6   /* Read data (+ MT,MFM,SK) */
#define CMD_SENSEI	0x08   /* Sense interrupt status  */
#define CMD_FORMAT	0x4D   /* Format track (+ MFM)    */
#define CMD_READID	0x4A   /* Read sector Id (+ MFM)  */
#define CMD_RECALIBRATE	0x07   /* Recalibrate             */
#define CMD_SEEK	0x0F   /* Seek track              */
#define CMD_VERSION 	0x10   /* Get FDC version         */

#define SAFESEEK	5      /* See recalibrate func    */

//fdc_t.flags
#define	DF_CHANGED 	0x01   /* Disk has been changed during last command */
#define DF_SPINUP   	0x02   /* Motor spinup time elapsed, OK to transfer */
#define DF_SPINDN   	0x03   /* Motor spindown time started               */

#define FDC_OK      	 0     /* successful completion */
#define FDC_ERROR   	-1     /* abnormal termination  */
#define FDC_NODISK  	-2     /* no disk in drive      */
#define FDC_TIMEOUT 	-3     /* timed out             */

#define FDC_READ	0
#define FDC_WRITE	1

#define MAX_DRIVES	2

typedef struct floppy_format_st
{
unsigned int    size;        /* Total number of sectors on disk    */
unsigned int    sec_per_trk; /* Number of sectors per track        */
unsigned int    heads;       /* Number of heads (sides)            */
unsigned int    tracks;      /* Number of tracks                   */
unsigned int    strech;      /* 1 if tracks are to be doubled      */
uint8_t         gap3;        /* GAP3 to use for reading/writing    */
uint8_t         rate;        /* Sector size and data transfer rate */
uint8_t         sr_hut;      /* Step Rate and Head Unload Time     */
uint8_t         gap3_fmt;    /* GAP3 to use for formatting         */
const char     *name;        /* Disk format name                   */
} floppy_format_t;

typedef struct drive_parameters_st
{
unsigned int   	cmos_type;  /* Drive type read from CMOS            */
unsigned int   	hlt;        /* Head Load Time (in SPECIFY format)   */
unsigned int   	spin_up;    /* Time needed for spinning up [us]     */
unsigned int   	spin_down;  /* Time needed for spinning down [us]   */
unsigned int   	sel_delay;  /* Select delay [us]                    */
unsigned int   	int_tmout;  /* Interrupt timeout [us]               */
unsigned int   	detect[8];  /* Autodetect formats                   */
unsigned int   	native_fmt; /* Native disk format for drive         */
const char     *name;       /* Name displayed for the drive type    */
} drive_parameters_t;

struct fdd_st;
typedef struct fdd_st fdd_t;
struct fdc_st;
typedef struct fdc_st fdc_t;

struct fdd_st
{
fdc_t         		  *fdc;       /* Controller of this drive       */
const drive_parameters_t  *dp;        /* Parameters for this drive      */
const floppy_format_t     *fmt;       /* Current format of the floppy   */
unsigned int              number;    /* Drive number per FDC            */
volatile unsigned         flags;     /* See the DF_* enums              */
unsigned int              track;     /* Current floppy track sought     */
};

struct fdc_st
{
uint16_t	base_port;
uint16_t 	flags;
uint8_t		result[7];
uint8_t		result_size;
uint8_t		sr0;
uint8_t		dor;
fdd_t		fdd[MAX_DRIVES];
};

typedef struct chs_st
{
unsigned int	cylinder;
unsigned int	head;
unsigned int	sector;
} chs_t;

//disk geometry
static const floppy_format_t floppy_formats[32] =
{
/*SIZE SPT HD TRK STR GAP3  RATE SRHUT GAP3F  NAME          NR DESCRIPTION   */
{    0,  0, 0,  0, 0, 0x00, 0x00, 0x00, 0x00, NULL    }, /*  0 no testing    */
{  720,  9, 2, 40, 0, 0x2A, 0x02, 0xDF, 0x50, "d360"  }, /*  1 360KB PC      */
{ 2400, 15, 2, 80, 0, 0x1B, 0x00, 0xDF, 0x54, "h1200" }, /*  2 1.2MB AT      */
{  720,  9, 1, 80, 0, 0x2A, 0x02, 0xDF, 0x50, "D360"  }, /*  3 360KB SS 3.5" */
{ 1440,  9, 2, 80, 0, 0x2A, 0x02, 0xDF, 0x50, "D720"  }, /*  4 720KB 3.5"    */
{  720,  9, 2, 40, 1, 0x23, 0x01, 0xDF, 0x50, "h360"  }, /*  5 360KB AT      */
{ 1440,  9, 2, 80, 0, 0x23, 0x01, 0xDF, 0x50, "h720"  }, /*  6 720KB AT      */
{ 2880, 18, 2, 80, 0, 0x1B, 0x00, 0xCF, 0x6C, "H1440" }, /*  7 1.44MB 3.5"   */
{ 5760, 36, 2, 80, 0, 0x1B, 0x43, 0xAF, 0x54, "E2880" }, /*  8 2.88MB 3.5"   */
{ 6240, 39, 2, 80, 0, 0x1B, 0x43, 0xAF, 0x28, "E3120" }, /*  9 3.12MB 3.5"   */
{ 2880, 18, 2, 80, 0, 0x25, 0x00, 0xDF, 0x02, "h1440" }, /* 10 1.44MB 5.25"  */
{ 3360, 21, 2, 80, 0, 0x1C, 0x00, 0xCF, 0x0C, "H1680" }, /* 11 1.68MB 3.5"   */
{  820, 10, 2, 41, 1, 0x25, 0x01, 0xDF, 0x2E, "h410"  }, /* 12 410KB 5.25"   */
{ 1640, 10, 2, 82, 0, 0x25, 0x02, 0xDF, 0x2E, "H820"  }, /* 13 820KB 3.5"    */
{ 2952, 18, 2, 82, 0, 0x25, 0x00, 0xDF, 0x02, "h1476" }, /* 14 1.48MB 5.25"  */
{ 3444, 21, 2, 82, 0, 0x25, 0x00, 0xDF, 0x0C, "H1722" }, /* 15 1.72MB 3.5"   */
{  840, 10, 2, 42, 1, 0x25, 0x01, 0xDF, 0x2E, "h420"  }, /* 16 420KB 5.25"   */
{ 1660, 10, 2, 83, 0, 0x25, 0x02, 0xDF, 0x2E, "H830"  }, /* 17 830KB 3.5"    */
{ 2988, 18, 2, 83, 0, 0x25, 0x00, 0xDF, 0x02, "h1494" }, /* 18 1.49MB 5.25"  */
{ 3486, 21, 2, 83, 0, 0x25, 0x00, 0xDF, 0x0C, "H1743" }, /* 19 1.74MB 3.5"   */
{ 1760, 11, 2, 80, 0, 0x1C, 0x09, 0xCF, 0x00, "h880"  }, /* 20 880KB 5.25"   */
{ 2080, 13, 2, 80, 0, 0x1C, 0x01, 0xCF, 0x00, "D1040" }, /* 21 1.04MB 3.5"   */
{ 2240, 14, 2, 80, 0, 0x1C, 0x19, 0xCF, 0x00, "D1120" }, /* 22 1.12MB 3.5"   */
{ 3200, 20, 2, 80, 0, 0x1C, 0x20, 0xCF, 0x2C, "h1600" }, /* 23 1.6MB 5.25"   */
{ 3520, 22, 2, 80, 0, 0x1C, 0x08, 0xCF, 0x2e, "H1760" }, /* 24 1.76MB 3.5"   */
{ 3840, 24, 2, 80, 0, 0x1C, 0x20, 0xCF, 0x00, "H1920" }, /* 25 1.92MB 3.5"   */
{ 6400, 40, 2, 80, 0, 0x25, 0x5B, 0xCF, 0x00, "E3200" }, /* 26 3.20MB 3.5"   */
{ 7040, 44, 2, 80, 0, 0x25, 0x5B, 0xCF, 0x00, "E3520" }, /* 27 3.52MB 3.5"   */
{ 7680, 48, 2, 80, 0, 0x25, 0x63, 0xCF, 0x00, "E3840" }, /* 28 3.84MB 3.5"   */
{ 3680, 23, 2, 80, 0, 0x1C, 0x10, 0xCF, 0x00, "H1840" }, /* 29 1.84MB 3.5"   */
{ 1600, 10, 2, 80, 0, 0x25, 0x02, 0xDF, 0x2E, "D800"  }, /* 30 800KB 3.5"    */
{ 3200, 20, 2, 80, 0, 0x1C, 0x00, 0xCF, 0x2C, "H1600" }  /* 31 1.6MB 3.5"    */

};

#define MS 1000 /*uS per mS*/
static const drive_parameters_t default_drive_params[] =
{
/* T HLT  spin_up  spin_dn  sel_d  int_tmt  Autodetect formats        Nat  Name     */
{ 0,  0, 1000*MS, 3000*MS, 20*MS, 3000*MS, { 7, 4, 8, 2, 1, 5, 3,10 }, 0, "unknown"                },
{ 1,  4, 1000*MS, 3000*MS, 20*MS, 3000*MS, { 1, 0, 0, 0, 0, 0, 0, 0 }, 1, "5.25\" DD, 360 KB"     },
{ 2,  8,  400*MS, 3000*MS, 20*MS, 3000*MS, { 2, 5, 6,23,10,20,12, 0 }, 2, "5.25\" HD, 1200 KB"    },
{ 3,  4, 1000*MS, 3000*MS, 20*MS, 3000*MS, { 4,22,21,30, 3, 0, 0, 0 }, 4, "3.5\" DD, 720 KB"      },
{ 4,  8,  400*MS, 3000*MS, 20*MS, 1500*MS, { 7, 4,25,22,31,21,29,11 }, 7, "3.5\" HD, 1440 KB"     },
{ 5, 15,  400*MS, 3000*MS, 20*MS, 3000*MS, { 7, 8, 4,25,28,22,31,21 }, 8, "3.5\" ED, 2880 KB AMI" },
{ 6, 15,  400*MS, 3000*MS, 20*MS, 3000*MS, { 7, 8, 4,25,28,22,31,21 }, 8, "3.5\" ED, 2880 KB"     }

};

#define DMA_CHANNEL 	2    /* Number of the DMA channel for floppy transfer */
#define DMA_PAGE    	0x81 /* Page register of that DMA channel             */
#define DMA_OFFSET  	0x04 /* Offset register of that DMA channel           */
#define DMA_LENGTH  	0x05 /* Length register of that DMA channel           */

fdc_t fdc;
int busy = 0;

void dma_init(unsigned int length, int write)
{
uint32_t page, offset;

if(write)
 printf("fdd: initializing DMA for reading %i bytes.\n", length);
else
 printf("fdd: initializing DMA for writing %i bytes.\n", length);

/* Calculate DMA page and offset */
page    = DMA_MEMORY_AREA >> 16;
offset  = DMA_MEMORY_AREA & 0xffff;
length -= 1; /* with DMA, if you want k bytes, you ask for k - 1 */

i_disable();                                       /* Disable IRQs            */
outportb(0x0A, DMA_CHANNEL | 4);                   /* Set channel mask bit    */
outportb(0x0C, 0);                                 /* Clear flip flop         */
outportb(0x0B, (write? 0x48 : 0x44) + DMA_CHANNEL);/* Mode (write+single+r/w) */
outportb(DMA_PAGE, page);                          /* Page                    */
outportb(DMA_OFFSET, offset & 0xFF);               /* Offset: low byte        */
outportb(DMA_OFFSET, offset >> 8);                 /* Offset: high byte       */
outportb(DMA_LENGTH, length & 0xFF);               /* Length: low byte        */
outportb(DMA_LENGTH, length >> 8);                 /* Length: high byte       */
outportb(0x0A, DMA_CHANNEL);                       /* Clear DMA mask bit      */
i_enable();                                        /* enable IRQs             */
}

int fdc_send(unsigned base_port, unsigned byte)
{
volatile unsigned msr;
unsigned          tmo;

for (tmo = 0; tmo < 128; tmo++)
 {
 msr = inportb(base_port + FDC_MSR);
 if ((msr & 0xC0) == 0x80)
  {
  outportb(base_port + FDC_DATA, byte);
  return FDC_OK;
  }
 inportb(0x80); /* delay */
 }
return FDC_TIMEOUT; /* write timeout */
}

int fdc_get(unsigned base_port)
{
volatile unsigned msr;
unsigned          tmo;

for (tmo = 0; tmo < 128; tmo++)
 {
 msr = inportb(base_port + FDC_MSR);
 if ((msr & 0xD0) == 0xD0)
  return inportb(base_port + FDC_DATA);
 inportb(0x80); /* delay */
 }
return FDC_TIMEOUT; /* read timeout */
}

int irq_signaled;

IRQ_HANDLER(fdc_interrupt)
{
irq_signaled = 1;
}

int wait_fdc(fdd_t *fdd)
{
/* Wait for IRQ6 handler to signal command finished */
volatile int irq_timeout = 0;
while(!irq_signaled && irq_timeout < FDC_IRQ_TIMEOUT)
 irq_timeout++;

/* Read in command result bytes while controller is busy */
fdd->fdc->result_size = 0;
while ((fdd->fdc->result_size < 7) && 
		(inportb(fdd->fdc->base_port + FDC_MSR) & 0x10))
 fdd->fdc->result[fdd->fdc->result_size++] = fdc_get(fdd->fdc->base_port);

irq_signaled = 0;
if(irq_timeout == FDC_IRQ_TIMEOUT)
 {
// printf("fdd: timeout\n");
 return FDC_TIMEOUT;
 }
return FDC_OK;
}

inline int is_motor_on(fdc_t *fdc, unsigned drive)
{
return fdc->dor & (1 << (drive + 4));
}

static void motor_on(fdd_t *fdd)
{
int n;
fdd->flags &= ~DF_SPINDN;
if (!is_motor_on(fdd->fdc, fdd->number))
 {
 if (fdd->flags & DF_SPINUP) 
  return;
 fdd->fdc->dor |= (1 << (fdd->number + 4));
 /* Select drive */
 fdd->fdc->dor = (fdd->fdc->dor & 0xFC) | fdd->number;
 outportb(fdd->fdc->base_port + FDC_DOR, fdd->fdc->dor);
 //TODO: make a real spinup delay!
 for(n = 0; n < 256000; n++);
 fdd->flags |= DF_SPINUP;
 }
}

void motor_down(fdd_t *fdd)
{
int n;
if (is_motor_on(fdd->fdc, fdd->number))
 {
 if (fdd->flags & DF_SPINDN) 
  return;
 fdd->flags |= DF_SPINDN;
 //TODO: make a real spindown delay!
 for(n = 0; n < 256000; n++);
 fdd->fdc->dor &= ~(1 << (fdd->number + 4));
 outportb(fdd->fdc->base_port + FDC_DOR, fdd->fdc->dor);
 fdd->flags &= ~(DF_SPINUP | DF_SPINDN);
 }
}

int fdc_disk_changed(fdd_t *fdd)
{
unsigned int changeline;
motor_on(fdd);
changeline = inportb(fdd->fdc->base_port + FDC_DIR) & 0x80;
motor_down(fdd);
return changeline || (fdd->flags & DF_CHANGED);
}

void recalibrate(fdd_t *fdd)
{
unsigned int k;
//printf("fdd: recalibrating\n");
for (k = 0; k < 13; k++)
 {
 fdc_send(fdd->fdc->base_port, CMD_RECALIBRATE);
 fdc_send(fdd->fdc->base_port, fdd->number);
 wait_fdc(fdd);
 /* Send a "sense interrupt status" command */
 fdc_send(fdd->fdc->base_port, CMD_SENSEI);
 fdd->fdc->sr0 = fdc_get(fdd->fdc->base_port);
 fdd->track    = fdc_get(fdd->fdc->base_port);
 if (!(fdd->fdc->sr0 & 0x10)) 
  break; /* Exit if Unit Check is not set */
 }
//printf("fdd: calibration result on drive %d: SR0=%x, track=%d\n",
//                fdd->number, fdd->fdc->sr0, fdd->track);
}

int fdc_seek(fdd_t *fdd, unsigned track)
{
if (fdd->track == track) return FDC_OK; /* Already there */
if (track >= fdd->fmt->tracks) return FDC_ERROR; /* Invalid track */

fdc_send(fdd->fdc->base_port, CMD_SEEK);
fdc_send(fdd->fdc->base_port, fdd->number);
fdc_send(fdd->fdc->base_port, track);
if (wait_fdc(fdd)) return FDC_TIMEOUT; /* Timeout */

/* Send a "sense interrupt status" command */
fdc_send(fdd->fdc->base_port, CMD_SENSEI);
fdd->fdc->sr0 = fdc_get(fdd->fdc->base_port);
fdd->track    = fdc_get(fdd->fdc->base_port);

/* Check that seek worked */
if ((fdd->fdc->sr0 != 0x20 + fdd->number) || (fdd->track != track))
 {
 printf("fdd: seek error on drive %d: SR0=%x, Track=%d Expected=%d\n",
                    fdd->number, fdd->fdc->sr0, fdd->track, track);
 return FDC_ERROR;
 }
//printf("fdd: seek result on drive %d: SR0=%x, Track=%d\n", fdd->number,
//                fdd->fdc->sr0, fdd->track);
return FDC_OK;
}

void specify(const fdd_t *fdd)
{
outportb(fdd->fdc->base_port + FDC_CCR, fdd->fmt->rate);
fdc_send(fdd->fdc->base_port, CMD_SPECIFY);
fdc_send(fdd->fdc->base_port, fdd->fmt->sr_hut);
fdc_send(fdd->fdc->base_port, fdd->dp->hlt << 1); /* Always DMA */
}

void reset_fdc(fdc_t *fdc)
{
unsigned int  k;
volatile int  irq_timeout = 0;

outportb(fdc->base_port + FDC_DOR, 0);   /* Stop the motor and disable IRQ/DMA*/
/* TODO: Add a small delay (20 us) to make older controllers more happy */
outportb(fdc->base_port + FDC_DOR, 0x0C);/* Re-enable IRQ/DMA & release reset */
fdc->dor = 0x0C;

while(!irq_signaled && irq_timeout < FDC_IRQ_TIMEOUT)
if (irq_timeout == FDC_IRQ_TIMEOUT)
 printf("fdd: timed out while waiting for FDC after reset\n");

/* FDC specs say to sense interrupt status four times */
for (k = 0; k < 4; k++)
 {
 /* Send a "sense interrupt status" command */
 fdc_send(fdc->base_port, CMD_SENSEI);
 fdc->sr0              = fdc_get(fdc->base_port);
 fdc->fdd[k % 2].track = fdc_get(fdc->base_port);
 }
irq_signaled = 0;
}

void reset_drive(fdd_t *fdd)
{
if (fdd->dp->cmos_type == 0)
 return;
fdd->flags   = 0;
fdd->track   = 0;
specify(fdd);
motor_on(fdd);
fdc_seek(fdd, SAFESEEK);
recalibrate(fdd);
motor_down(fdd);
fdd->flags |= DF_CHANGED; /* So that fdc_log_disk can be issued */
}

int fdc_xfer(fdd_t *fdd, const chs_t *chs, unsigned int num_sectors, int op)
{
int timeout = 0;
//printf("fdd: fdc_xfer: C=%d, H=%d, S=%d, n=%d, op=%d\n", 
//    chs->cylinder, chs->head, chs->sector, num_sectors, op);

/* Wait for motor spin quickly enough */
while(!(fdd->flags & DF_SPINUP));

/* Send read/write command */
if (op == FDC_READ)
 {
 dma_init(512 * num_sectors, 0); /* Read */
 fdc_send(fdd->fdc->base_port, CMD_READ);
 }
else
 {
 dma_init(512 * num_sectors, 1); /* Write */
 fdc_send(fdd->fdc->base_port, CMD_WRITE);
 }

//printf("fdd: reading\n");
fdc_send(fdd->fdc->base_port, (chs->head << 2) | fdd->number);
fdc_send(fdd->fdc->base_port, chs->cylinder);
fdc_send(fdd->fdc->base_port, chs->head);
fdc_send(fdd->fdc->base_port, chs->sector);
fdc_send(fdd->fdc->base_port, 2); /* 512 bytes per sector */
fdc_send(fdd->fdc->base_port, fdd->fmt->sec_per_trk);
fdc_send(fdd->fdc->base_port, fdd->fmt->gap3);
fdc_send(fdd->fdc->base_port, 0xFF); /* DTL (bytes to transfer) = unused */

/* Wait for completion and produce exit code */
if (wait_fdc(fdd) == FDC_TIMEOUT)
 {
 reset_fdc(fdd->fdc);
 reset_drive(fdd);
 return FDC_TIMEOUT;
// printf("fdc: timeout...\n");
 }
if ((fdd->fdc->result[0] & 0xC0) == 0) 
 return FDC_OK;
return FDC_ERROR;
}

int fdc_read(fdd_t *fdd, const chs_t *chs, 
		    uint8_t *buffer, unsigned int num_sectors)
{
unsigned tries, tmout = 0;
int      res = FDC_ERROR;

if (fdd->dp->cmos_type == 0)
 return FDC_ERROR; /* Drive not available */
 
/* TODO: Place a timeout for Busy check */
while (busy); /* Wait while the floppy driver is already busy: BUSY WAIT! */
busy = 1;
motor_on(fdd);
specify(fdd);
for (tries = 0; tries < 3; tries++)
 {
 /* Move head to right track */
 if(fdc_seek(fdd, chs->cylinder) == FDC_OK)
  {
  /* If changeline is active, no disk is in drive */
  if(inportb(fdd->fdc->base_port + FDC_DIR) & 0x80)
   {
   printf("fdd: fdc_read: no disk in drive\n");
   res = FDC_NODISK;
   break;
   }
  res = fdc_xfer(fdd, chs, num_sectors, FDC_READ);
  if (res == FDC_OK) 
   break;
  }
 else res = FDC_ERROR; /* Seek error */
 if (tries != 2) 
  recalibrate(fdd); /* Error, try again... */
 }
/* Copy data from the DMA buffer into the caller's buffer */
if((res == FDC_OK) && buffer)
 memcpy(DMA_MEMORY_AREA, buffer, num_sectors * 512);
motor_down(fdd);
busy = 0;
return res;
}

int probe_format(fdd_t *fdd, unsigned format)
{
chs_t chs = { 0, 0, 0 };
int res;

fdd->fmt = &floppy_formats[format];
chs.sector = fdd->fmt->sec_per_trk;
res = fdc_read(fdd, &chs, NULL, 1);
//printf("fdd: probe_format: fdc_read returned %i\n", res);
if (res < 0) 
 return res;
//printf("%s format detected\n", fdd->fmt->name);
return FDC_OK;
}

int fdc_log_disk(fdd_t *fdd)
{
unsigned int k;
if (fdd->dp->cmos_type == 0) 
 return FDC_ERROR; /* Drive not available */
fdd->fmt = &floppy_formats[fdd->dp->native_fmt];
reset_fdc(fdd->fdc);
reset_drive(fdd);
/* If changeline is active, no disk is in drive */
motor_on(fdd);
if (inportb(fdd->fdc->base_port + FDC_DIR) & 0x80)
 {
 printf("fdd: fdc_log_disk: no disk in drive\n");
 motor_down(fdd);
 return FDC_NODISK;
 }
fdd->flags &= ~DF_CHANGED;
for (k = 0; k < 8; k++)
 if(probe_format(fdd, fdd->dp->detect[k]) == 0)  
  {
  //TODO: fix it
  /* Finetune the probed format with boot sector informations */
//  int res = floppy_bootsector(fdd, floppy_formats, 32);
/*  if (res >= 0)*/ fdd->fmt = &floppy_formats[7]; //H1440
  return FDC_OK;
  }
motor_down(fdd);
fdd->flags |= DF_CHANGED; /* Must log the disk again the next time */
return FDC_ERROR; /* If we arrive here, the autodetection failed */
}

/* Initializes a drive to a known state */
static int setup_drive(fdc_t *fdc, unsigned int drive, unsigned int cmos_type)
{
fdc->fdd[drive].fdc       = fdc;
fdc->fdd[drive].dp        = &default_drive_params[0];

fdc->fdd[drive].number    = drive;
fdc->fdd[drive].flags     = 0;
fdc->fdd[drive].track     = 0;

if ((cmos_type > 0) && (cmos_type <= 6))
 {
 printf("fdd: drive %d set to %s\n", drive, default_drive_params[cmos_type].name);
 fdc->fdd[drive].dp        = &default_drive_params[cmos_type];
 fdc->fdd[drive].flags     = DF_CHANGED;
 }
return FDC_OK;
}

void fdd_blkdev_driver(int minor, blkdev_operation_t op)
{
if(op.cmd == BLKDEV_INFO)
 {
 printf("fdd_blkdev_driver: info request (minor = %i)\n", minor);
 blkdev_info_t *info = op.buf;
 //TODO: make it actual information
 info->block_size = 512;
 info->block_count = 2880;
 }
else if(op.cmd == BLKDEV_READ)
 {
 printf("fdd_blkdev_driver: read request (minor=%i, from=%i, cnt=%i)\n", 
					minor, op.start_block, op.block_count);
 fdd_t fdd = fdc.fdd[minor];
 chs_t chs;
 
 unsigned int lba = op.start_block;
 unsigned int hpc = 2; //heads per cylinder;
 chs.cylinder = lba / (hpc * fdd.fmt->sec_per_trk);
 unsigned int temp = lba % (hpc * fdd.fmt->sec_per_trk);
 chs.head = temp / fdd.fmt->sec_per_trk;
 chs.sector = temp % fdd.fmt->sec_per_trk + 1;
 
 if(fdc_read(&fdc.fdd[minor], &chs, op.buf, op.block_count) == FDC_OK)
  return 0;
 }
return -1;
}

void init_fdd(nsentry_t *dev)
{
int stat;
nsentry_t *fd0, *fd1, *raw;

printf("fdd: initializing...\n");

i_install(FDC_IRQ, &fdc_interrupt, RING0 | INTERRUPT_GATE);
i_enable();

printf("fdd: installed interrupt\n");

//TODO: check FDC

outportb(CMOS_REG, 0x10);
stat = inportb(CMOS_DAT);

assert(request_system_memory_area(DMA_MEMORY_AREA, 0x10, 0) != -1);
			//64k DMA page
printf("fdd: allocated a DMA buffer at %X\n", DMA_MEMORY_AREA);

fdc.base_port = FDC_PRI_IO;

if((stat >> 4) != 0x0)
 {
 printf("fdd: detected a floppy drive at position 0\n");
 fd0 = add_nsentry(dev, create_nsentry(NSE_CONTAINER), "fd0");
 raw = add_nsentry(fd0, create_nsentry(NSE_OBJECT), "raw");
 raw->nsstatus.vfsid = VFS_DEVFS;
 raw->nsstatus.ident =  get_ident(DEVFS_IDENT_FLOPPY, 0);
 raw->nsstatus.size = 0x1000;
 }

if((stat & 0x0F) == 0x04)
 {
 printf("fdd: detected a floppy drive at position 1\n");
 fd1 = add_nsentry(dev, create_nsentry(NSE_CONTAINER), "fd1");
 raw = add_nsentry(fd1, create_nsentry(NSE_OBJECT), "raw");
 raw->nsstatus.vfsid = VFS_DEVFS;
 raw->nsstatus.ident =  get_ident(DEVFS_IDENT_FLOPPY, 1);
 raw->nsstatus.size = 0x1000;
 }
 
setup_drive(&fdc, 0, stat >> 4);
setup_drive(&fdc, 1, stat & 0x0F);

register_blkdev_driver(DEVFS_IDENT_FLOPPY, "fdd", &fdd_blkdev_driver);

busy = 0;
}