/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

/*
  IDE HDD driver
  
  TODO:
  - write support
  - DMA mode
  - move partitioning stuff out of here
  (partitioning really have almost nothing to do with hdd driver :)
  
*/

#include <tyros.h>
#include <ntx.h>
#include <ntxbase.h>
#include <filesystems.h>
#include <devfs.h>

#define ata_timeout 10000

nsentry_t *dev;

void scan_devices();
char detect_device(int atanum, int devnum);

#define DEVTYPE_ATA 0
#define DEVTYPE_ATAPI 1

//cache size per device in PAGES
#define CACHE_SIZE 0x20 //128k

typedef struct disk_cache_struct {
  fsize_t size;
  fsize_t addr;
  char *data;
} disk_cache_t;

typedef struct partition_struct {
  unsigned int start;
  unsigned int size;
} partition_t;

typedef struct ata_device_struct {
  partition_t partition[32];
} ata_device_t;


disk_cache_t disk_cache[2][2];

ata_device_t ata_device[2][2];

IRQ_HANDLER(ata0_interrupt)
{
  printf("hdd: ata0 interrupt\n");
}

IRQ_HANDLER(ata1_interrupt)
{
  printf("hdd: ata1 interrupt\n");
}


int init_caching(int atanum, int devnum)
{
  disk_cache[atanum][devnum].data = alloc_pmem(CACHE_SIZE);
}

/*
  transforms ATA channel number, drive number and partition number into single integer
*/
unsigned int get_ata_minor (int atanum, int devnum, int pnum)
{
  unsigned int minor;
  
  minor = pnum | (devnum << 8) | (atanum << 12);
  
  return minor;
}
/*
  reverses the result of previous func
*/
int unpack_ata_minor(unsigned int minor, int *atanum, int *devnum, int *pnum)
{
  *atanum = (minor & 0x0000F000) >> 12;
  *devnum = (minor & 0x00000F00) >> 8;
  *pnum =   (minor & 0x000000FF);
}


/*
  read drive/partiton:
  minor:  packed atanum/devnum/pnum
  start:  starting offset (in bytes)
  size:   total size (in bytes)
*/
int read_ata (int minor, fsize_t start, fsize_t size, void *buffer)
{
  int atanum, devnum, pnum;
  fsize_t realstart;
  partition_t *part;

  //  printf("reading ata. Minor: %i\n", minor);
  
  unpack_ata_minor(minor, &atanum, &devnum, &pnum);
  
  //  printf("atanum: %i, devnum: %i, pnum: %i\n", atanum, devnum, pnum);

  if(pnum == 0) {
    return read_hdd(atanum, devnum, start, size, buffer);
  } else {
    //    printf("reading logical partition %i\n", pnum);
    part = &ata_device[atanum][devnum].partition[pnum];
    realstart = part->start*512 + start;
    //    printf("realstart: %X\n", realstart);
    return read_hdd(atanum, devnum, realstart, size, buffer);
  }

}

/*
  the same as previous func BUT doesn't deal with partitions and read the 'raw' drive
*/
int read_hdd (int atanum, int devnum, fsize_t start, fsize_t size, void *buffer)
{
  int prefix, postfix;
  int sector, secs;
  int i;
  disk_cache_t *cache;

  prefix = start%512;
  postfix = size%512;

  sector = start/512;
  secs = size/512 + 1;

  //  printf("Start: %X, size: %X\n", (unsigned int) start, (unsigned int) size);
  //  printf("Sector start: %i\n", sector);
  //  printf("Sectors count: %i\n", secs);
  //  printf("Prefix: %i, Postfix: %i\n", prefix, postfix);
  
  
  cache = &disk_cache[atanum][devnum];
  
  //  printf("cache->addr = %X, cache->size = %X\n", (unsigned int) cache->addr,
  //	 (unsigned int) cache->size);
  
  if((start >= cache->addr) && 
     (cache->addr + cache->size) > start) {
    //we've hit the cache
    printf("cache hit\n");
    for(;;);
    
  } else {
    //cache miss :(
    //    printf("cache miss\n");
    
    //reload cache
    
    if(size > CACHE_SIZE*PAGE_SIZE) {
      printf("read request size exceeds cache size!\n");
      for(;;);
    } else {
      //      printf("reading %i sectors into cache\n", secs);
      for(i = 0; i < secs; i++) {
	read_sector (atanum, devnum, sector + i, cache->data + i*512);
      }
      memcpy(buffer, cache->data + prefix, (unsigned int) size);
      return size;
    }
    
  }
  
}

/*
  init hdd driver
*/
void init_hdd(nsentry_t *dev_nsentry)
{
  dev = dev_nsentry;
  
  scan_devices();
  
  i_install (0x20 + 14, &ata0_interrupt, RING0 | INTERRUPT_GATE);
  i_install (0x20 + 15, &ata1_interrupt, RING0 | INTERRUPT_GATE);


  //  ata_test();
}


/*
  init partition
  
  dir:   directory where partition files will appear
  entry: pointer to partition descriptor (from MBR)
  num:   logical number of partition
*/
void init_entry (int atanum, int devnum, nsentry_t *dir, char *entry, int num)
{
  unsigned char id, state;
  nsentry_t *pns, *raw;
  char *number;
  partition_t *part;
  char *start;
  char *size;
  char smbr[0x210];
  int exstart, exsize;
  int eps; //extended partition start

  id = entry[4];
  state = entry[0];
  
  part = &ata_device[atanum][devnum].partition[num];
  
  if(id != 0) {
    printf("hdd: partition %i: id: %x, state: %s\n", num, id, 
	   state == 0x80? "active" : "inactive");

    part->start = * (unsigned int *) (&entry[0x08]);
    part->size = * (unsigned int *) (&entry[0x0C]);

    if(id != 5 && id != 0xF) {
      
      printf("hdd: start: %i, size: %i\n", part->start, part->size);
    
      number = malloc(32);
      size = malloc(32);
      start = malloc(32);
      
      sprintf(number, "%i", num);
      
      pns = add_nsentry (dir, 
			 create_nsentry(NSE_CONTAINER),
			 number);
      
      sprintf(start, "%i\n", part->start);
      sprintf(size, "%i\n", part->size);
      
      add_static_nsentry (pns, "start", start);
      add_static_nsentry (pns, "size", size);
      
      raw = add_nsentry (pns,
			 create_nsentry(NSE_OBJECT),
			 "raw");
      
      raw->nsstatus.vfsid = VFS_DEVFS;
      raw->nsstatus.ident =  get_ident(DEVFS_IDENT_ATA, get_ata_minor(atanum, devnum, num));
      raw->nsstatus.size = 0x100000;
      
    } else {
      printf("hdd: extended partition found\n");

      eps = part->start;
      exstart = 0;
      exsize = part->size;
      
      //logical partitions are numbered beginning with 5
      num = 5;
      
      do {
	  
	//	printf("Start: %i, size: %i\n", exstart, exsize);
	
	read_sector (atanum, devnum, eps + exstart, smbr);
	
	part = &ata_device[atanum][devnum].partition[num];
	
	part->start = * (int *) &smbr[0x1BE + 0x08] + eps + exstart;
	part->size = * (int *) &smbr[0x1BE + 0x0C];
	

	//	printf("extended: type: %x, rstart: %i\n", smbr[0x1BE + 0x04], part->start);

	number = malloc(32);
	size = malloc(32);
	start = malloc(32);
	
	sprintf(number, "%i", num);
	
	pns = add_nsentry (dir, 
			   create_nsentry(NSE_CONTAINER),
			   number);
	
	sprintf(start, "%i\n", part->start);
	sprintf(size, "%i\n", part->size);
	
	add_static_nsentry (pns, "start", start);
	add_static_nsentry (pns, "size", size);
	
	raw = add_nsentry (pns,
			   create_nsentry(NSE_OBJECT),
			   "raw");
	
	raw->nsstatus.vfsid = VFS_DEVFS;
	raw->nsstatus.ident =  get_ident(DEVFS_IDENT_ATA, get_ata_minor(atanum, devnum, num));
	raw->nsstatus.size = 0x100000;
	
	
	exstart = * (int *) &smbr[0x1CE + 0x08];
	
	//	printf("Next partition: %i, type: %x\n", exstart, smbr[0x1CE + 0x04]);
	  
	num++;
      } while(smbr[0x1CE + 0x04] != 0);
      
      
      
      
    }
    
  }  
}

/*
  initializes all partitions of given drive

  dir:      directory where partition files will appear
  devtype:  DEVTYPE_ATA/DEVTYPE_ATAPI
  infobuf:  pointer to drive information buffer
            (as returned by IDENTIFY DRIVE command)
*/
void init_partitions (int atanum, int devnum, 
		      nsentry_t *dir, int devtype,
		      char *infobuf)
{
  char mbr[0x200];
  char entry1[0x10];
  char entry2[0x10];
  char entry3[0x10];
  char entry4[0x10];

  printf("hdd: initializing partitions for %i/%i\n", atanum, devnum);
  read_sector (atanum, devnum, 0, mbr);
  
  memcpy(entry1, mbr + 0x1BE, 0x10);
  memcpy(entry2, mbr + 0x1CE, 0x10);
  memcpy(entry3, mbr + 0x1DE, 0x10);
  memcpy(entry4, mbr + 0x1FE, 0x10);
  
  printf("hdd: initializing mbr entries...\n");

  init_entry(atanum, devnum, dir, entry1, 1);
  init_entry(atanum, devnum, dir, entry2, 2);
  init_entry(atanum, devnum, dir, entry3, 3);
  init_entry(atanum, devnum, dir, entry4, 4);
 
}


/*
  add the given drive into system (and init it)
*/
void add_ata_device(int atanum, int devnum, int devtype, char *infobuf)
{
  nsentry_t *device_dir, *raw;
  char *devname;
  unsigned short *buffer;
  int i;
  char *model, *serial;

  
  devname = malloc(16);
  sprintf(devname, "ata%i%i",atanum, devnum); 

  device_dir = add_nsentry(dev, create_nsentry(NSE_CONTAINER), devname);
  
  buffer = infobuf;
  
  
  serial = malloc(32);
  model = malloc(64);
  
  *serial = *model = 0;
  
  for(i = 27; i < 47; i++) {
    model[ (i - 27)*2] = (buffer[i] & 0xFF00) >> 8;
    model[ (i - 27)*2 + 1] = buffer[i] & 0x00FF;
  }
  
  model[40] = '\n';
  model[41] = 0;

  add_static_nsentry(device_dir, "model", model);
  

  for(i = 10; i < 20; i++) {
    serial[ (i - 10)*2] = (buffer[i] & 0xFF00) >> 8;
    serial[ (i - 10)*2 + 1] = buffer[i] & 0x00FF;
  }

  serial[20] = '\n';
  serial[21] = 0;

  add_static_nsentry(device_dir, "serial", serial);
  
  raw = create_nsentry(NSE_OBJECT);
  
  raw->nsstatus.vfsid = VFS_DEVFS;
  raw->nsstatus.ident = DEVFS_IDENT_ATA | get_ata_minor(atanum, devnum, 0);
  raw->nsstatus.size = hal.ramdisk_size;  
  
  add_nsentry(device_dir, raw, "raw");

  init_caching(atanum, devnum);

  if(devtype == DEVTYPE_ATA) {
    init_partitions(atanum, devnum, device_dir, devtype, infobuf);
  } else if(devtype == DEVTYPE_ATAPI) {
    
  } else {
    panic("hdd: ATA device type neither ATA nor ATAPI\n");
  }

}

#define ATA0 0x1F0
#define ATA1 0x170

#define ATAREG_DR 0
#define ATAREG_ER 1
#define ATAREG_FR 1
#define ATAREG_SC 2
#define ATAREG_SN 3
#define ATAREG_CL 4
#define ATAREG_CH 5
#define ATAREG_DH 6
#define ATAREG_DEV 6
#define ATAREG_SR 7
#define ATAREG_CR 7

#define SR_BSY 0x80
#define SR_DRDY 0x40
#define SR_DF 0x20
#define SR_SERV 0x10
#define SR_DRQ 0x8
#define SR_CHK 0x1
#define SR_ERR 0x1




char device_exists (int atanum, int devnum)
{
  char sr;
  int channel;
  char device;

  if(atanum == 0) {
    channel = ATA0;
  } else if (atanum == 1){
    channel = ATA1;
  } else {
    return 0;
  }
  
  if(devnum == 0) {
    device = 0xE0;
  } else if(devnum == 1) {
    device = 0xF0;
  } else {
    return 0;
  }

  do {
  sr = inportb(channel | ATAREG_SR);
  } while (! (!(sr & SR_BSY) && !(sr & SR_DRQ)));
  
  outportb (channel | ATAREG_DEV, device);
  
  do {
    sr = inportb(channel | ATAREG_SR);
  } while (! (!(sr & SR_BSY) && !(sr & SR_DRQ)));

  
  if(sr & SR_CHK) {
    return 0;
  } else {
    return 1;
  }
}


/*
  scans ATA0 and ATA1 channels for drives
*/
void scan_devices()
{
  int i, j;
  for(i = 0; i < 1; i++) {  //only ata0 bus - fixme
    for(j = 0; j < 2; j++) {
      if(device_exists(i,j)) {
	detect_device(i,j);
      }
    }
  }
}

/*
  get channel starting port area from its number
*/
int get_channel (int atanum)
{
  if(atanum == 0) {
    return ATA0;
  } else if (atanum == 1){
    return ATA1;
  } else {
    panic("hdd: get_channel: atanum = %i!\n", atanum);
  }
}

/*
  get drive code from its number
*/
int get_device (int devnum)
{
  if(devnum == 0) {
    return 0xE0;
  } else if(devnum == 1) {
    return 0xF0;
  } else {
    panic("hdd: get_device: devnum = %i!\n", devnum);
  }
}

/*
  reads specified sector
*/
int read_sector (int atanum, int devnum, unsigned int lba, void *buffer)
{
  unsigned int channel;
  unsigned char device;
  unsigned char sr;
  

  channel = get_channel(atanum);
  device = get_device(devnum);

  do
   {
   sr = inportb(channel | ATAREG_SR);
   }
  while( sr & SR_BSY ||
	 sr & SR_DRQ);
  

  outportb(channel | ATAREG_DEV, device | (lba & 0x0F000000) >> 24);

  do
   {
   sr = inportb(channel | ATAREG_SR);
   }
  while( sr & SR_BSY ||
	 sr & SR_DRQ);

  outportb(channel | ATAREG_SC, 1);
  outportb(channel | ATAREG_SN, (lba & 0x000000FF));
  outportb(channel | ATAREG_CL, (lba & 0x0000FF00) >> 8);
  outportb(channel | ATAREG_CH, (lba & 0x00FF0000) >> 16);
  outportb(channel | ATAREG_CR, 0x20);

  while( (inportb(channel | ATAREG_SR) & 8) == 0){
    //    printf("Waiting....\n");
  }

  asm("rep insw"::"D"(buffer),"c"(512/2),"d"(channel | ATAREG_DR));
  
  //  printf("reading ok!\n");
  
}


/*
  Wait until specified channel is ready
*/
int wait_channel (int channel)
{
  int i;
  char sr;

  i = 0;
  do {
    i++;
    sr = inportb (channel | ATAREG_SR);
  }  while( (sr & SR_BSY || sr & SR_DRQ) && i < ata_timeout);

  if(i >= ata_timeout) {
    return 0;
  } else {
    return 1;
  }
}

char detect_device(int atanum, int devnum)
{
  int i;
  char sr;
  int channel;
  char device;
  short buffer[256];
  int type;

  if(atanum == 0) {
    channel = ATA0;
  } else if (atanum == 1){
    channel = ATA1;
  } else {
    return 0;
  }
  
  if(devnum == 0) {
    device = 0xE0;
  } else if(devnum == 1) {
    device = 0xF0;
  } else {
    return 0;
  }

  i = 0;
  do {
    i++;
    sr = inportb (channel | ATAREG_SR);
  }  while( (sr & SR_BSY || sr & SR_DRQ) && i < ata_timeout);

  outportb (channel | ATAREG_DEV, device);
  
  
  i = 0;
  do {
    i++;
    sr = inportb (channel | ATAREG_SR);
  } while( (sr & SR_BSY || sr & SR_DRQ) && i < ata_timeout);
  
  
  outportb (channel | ATAREG_CR, 0xEC);
  
  do {
    sr = inportb (channel | ATAREG_SR);
  } while(sr & SR_BSY);
  
  if( !(sr & SR_CHK)) {
    type = 0;

  } else {
   
    i = 0;
    do {
      i++;
      sr = inportb (channel | ATAREG_SR);
    } while( (sr & SR_BSY) && i < ata_timeout);
    
    
    outportb (channel | ATAREG_CR, 0xA1);
    
    i = 0;
    do {
      i++;
      sr = inportb (channel | ATAREG_SR);
    } while( (sr & SR_BSY) && i < ata_timeout); 
    
    if(sr & SR_CHK) {
      printf("hdd: no device found on %i/%i\n", atanum, devnum);
      return 0;
    } else {
      type = 1;
    }
  }
  
  for(i = 0; i < ata_timeout; i++) {
    sr = inportb(channel | ATAREG_SR);
    if(sr & SR_DRQ){
      break;
    }
  }
  
  if(i >= ata_timeout) {
    printf("hdd: scanning %i/%i: timeout\n", atanum, devnum);
    return -1;
  }
  
  
  for(i = 0; i < 256; i++) {
    buffer[i] = inportw(channel | ATAREG_DR);
  }
  
  if(type == 0) {
    printf("hdd: ATA device detected on ATA%i/%i\n", atanum, devnum);
    add_ata_device(atanum, devnum, DEVTYPE_ATA, buffer);
  } else {
    printf("hdd: ATAPI device detected on ATA%i/%i\n", atanum, devnum);
    add_ata_device(atanum, devnum, DEVTYPE_ATAPI, buffer);
  }
  
}
