/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <stdarg.h>

#define VIDEO_RAM 0xb8000
#define VIDEO_WIDTH 80
#define VIDEO_HEIGHT 25

void load_image_fat32();
void load_kernel_fat12();
void load_cluster_chain (int cluster, void *dest);
void copy_cluster (int cluster, void *dest);
void init_tty();
void update_cursor(unsigned short position);
inline char inportb(unsigned short port);
inline void outportb(unsigned short port, unsigned char value);
void clear();
void puthexi(unsigned int dword);
void puthex(unsigned char byte);
void puthexd(unsigned char digit);
void dvprintf(const char *fmt, va_list args);

typedef unsigned int Elf32_Addr;
typedef unsigned short Elf32_Half;
typedef unsigned int Elf32_Off;
typedef signed int Elf32_SWord;
typedef unsigned int Elf32_Word;

#define EI_NIDENT 16

#define SHT_STRTAB 3

#define SHF_WRITE 0x1
#define SHF_ALLOC 0x2
#define SHF_EXECINSTR 0x4


struct elf32_hdr {
  unsigned char e_ident[EI_NIDENT];
  Elf32_Half e_type;
  Elf32_Half e_machine;
  Elf32_Word e_version;
  Elf32_Addr e_entry;
  Elf32_Off e_phoff;
  Elf32_Off e_shoff;
  Elf32_Word e_flags;
  Elf32_Half e_ehsize;
  Elf32_Half e_phentsize;
  Elf32_Half e_phnum;
  Elf32_Half e_shentsize;
  Elf32_Half e_shnum;
  Elf32_Half e_shstrndx;
};


struct elf32_phdr {
  Elf32_Word p_type;
  Elf32_Off p_offset;
  Elf32_Addr p_vaddr;
  Elf32_Addr p_paddr;
  Elf32_Word p_filesz;
  Elf32_Word p_memsz;
  Elf32_Word p_flags;
  Elf32_Word p_align;
};

struct elf32_shdr {
  Elf32_Word sh_name;
  Elf32_Word sh_type;
  Elf32_Word sh_flags;
  Elf32_Addr sh_addr;
  Elf32_Off sh_offset;
  Elf32_Word sh_size;
  Elf32_Word sh_link;
  Elf32_Word sh_info;
  Elf32_Word sh_addralign;
  Elf32_Word sh_entsize;
};


#define PT_LOAD 1

int ctor_size;

int tty_cursor;
int tty_attribute;


int load_elf (unsigned char *bin)
{
  int i;
  int s;
  struct elf32_hdr *EH = (struct elf32_hdr *)bin;
  struct elf32_phdr *EPH;
  struct elf32_shdr *ESH, *strtab;
  int back, back2, ctor;
  void (*ctor_entry)();
  
  EPH = (struct elf32_phdr *)(bin + EH->e_phoff);
  ESH = (struct elf32_shdr *)(bin + EH->e_shoff);
  printf("Program header table is found at %X\n", EPH);
  printf("Sections header table is found at %X\n", ESH);

  printf("Number of sections: %i/%i\n", EH->e_phnum, EH->e_shnum);
  
  while (EH->e_phnum--) {
    if (EPH->p_type == PT_LOAD) {
      memcpy (EPH->p_vaddr, bin + EPH->p_offset, EPH->p_filesz);
      printf("Loading section (%X, size %i)\n", bin + EPH->p_offset, EPH->p_filesz);
    }
    EPH = (struct elf32_phdr *)((unsigned char *)EPH + EH->e_phentsize);
  }
  
  
  return ctor_entry;
}

int ATANUM;
int DEVNUM;


void main(int bootdata)
{
  void (*kernel)(int);
  char *video = VIDEO_RAM;
  int load_from;
  
  load_from = * (int *) (bootdata + 8);
  
  * (int *) (bootdata + 12) = 0x110000;

  init_tty();
  clear();

  printf("Neutronix Kernel Loader is Ready\n");

  switch(load_from) {
  case 1:
    ATANUM = 0;
    DEVNUM = 0;
    load_image_fat32();
    break;

  case 2:
    ATANUM = 0;
    DEVNUM = 1;
    load_image_fat32();
    break;
  }
  
  load_kernel_fat12();
  

  kernel = 0x200000;
  
  
  
  kernel(bootdata);

  for(;;);
}

#define RAMDISK_START 0x400000

typedef struct fat_info {
  char oem_name[0x10];
  int bytes_per_sec;
  int sec_per_clust;
  int reserved_sectors;
  int numfats;
  int root_entries;
  int total_sectors;
  int media;
  int fat_size;
  int sec_per_track;
  int num_heads;
  int hidden_sectors;
  int total_sectors32;
  int root_cluster;
} fat_info_t;


char mbr[0x210];
char bpb[0x210];
fat_info_t table;
unsigned int cstart, csize;
unsigned char pcode;

unsigned int get_fat_value (int cluster)
{
  int sector[0x90];
  unsigned int ts;//target sector
  unsigned int tv;
  unsigned int fatstart;

  fatstart = table.reserved_sectors;

  ts = cstart + fatstart + cluster/128;
  tv = cluster%128;
  
  read_sector(ATANUM,DEVNUM, ts, sector);
  
  return sector[tv];
}

int load_cluster_chain32(int cluster, void *buffer)
{
  int nextcluster;
  int n;

  printf("Loading cluster chain %X to %X\n", cluster, buffer);
  
  n = 0;
  
  for(;;){
    nextcluster = get_fat_value(cluster);
    load_cluster32(cluster, buffer + n*table.sec_per_clust*512);
    putchar('>');
    n++;
    cluster = nextcluster;
    if(cluster >= 0x0FFFFFF8) {
      break;
    }
  }

  printf("cluster chain is loaded\n");
}

int load_cluster32(int cluster, void *buffer)
{
  int start;//first sector
  int i;

  //  printf("Loading cluster %X to %X\n", cluster, buffer);
  
  start = table.fat_size*table.numfats+table.reserved_sectors+(cluster - 2)*table.sec_per_clust;
  
  for(i = 0; i < table.sec_per_clust; i++) {
    read_sector(ATANUM,DEVNUM,cstart + start + i, buffer + i*512);
  }
  
}

unsigned int get_cluster32 (char *filename)
{
  char rootdir[0x10000]; //64kb in case someone has such big clusters
  int cluster;
  int nextcluster;
  int i;
  int eps; //entries per cluster
  char name[12];
  int found;
  int fclust;//found cluster
  unsigned int hword = 0, lword = 0;

  name[11] = 0;

  cluster = table.root_cluster;

  eps = table.sec_per_clust*16;
  

 
  found = 0;
  

  for(;;) {
    load_cluster32(cluster, rootdir);
    nextcluster = get_fat_value(cluster);
    for(i = 0; i < eps; i++) {
      if(rootdir[i*32] != 0) {
	
	if(rootdir[i*32] != 0xE5) {
	  memcpy(name, &rootdir[i*32], 11);
	  if(!strcmp(name, filename)) {
	    found = 1;
	    hword = (unsigned int) * (short *) &rootdir[i*32 + 0x14] & 0x0000FFFF;
	    lword =  (unsigned int) * (short *) &rootdir[i*32 + 0x1A] & 0x0000FFFF;
	      
	    printf("high word: %X\n", hword);
	    printf("low word: %X\n", lword);
	    
	    fclust = (hword << 16) | lword;
	    
	    printf("fclust: %X\n", fclust);
	    
	    break;
	  }
	}    
	
      } else {
	break;
      }
    }
    cluster = nextcluster;
    if(found) {
      break;
    }
    if(cluster >= 0x0FFFFFF8) {
      break;
    }
  }
  
  if(found) {
    printf("file has been found! Cluster: %X\n", fclust);
    return fclust;
  } else {
    printf("CAN'T FIND %s!\n", filename);
    for(;;);
  }

}

void load_image_fat32()
{
  int fcluster;

  printf("Loading ramdisk image from FAT32 hdd...\n");
  
  read_sector (ATANUM,DEVNUM,0,mbr);
  
  cstart = * (int *) (mbr + 0x1BE + 0x08);
  csize = * (int *) (mbr + 0x1BE + 0x0C);
  pcode = * (mbr + 0x1BE + 0x04);

  printf("C:\\ starts at sector %i and contains %i sectors (%i mb)\n", cstart, csize, (csize/2096));
  
  printf("Partition code: %x\n", pcode);

  if(pcode != 0x0B && pcode != 0x0C) {
    printf("Partition code (%x) reports non-FAT32 structure!\n", pcode);
  }

  read_sector(ATANUM,DEVNUM,cstart, bpb);
  
  fill_fat_info(&table, bpb);
  
  printf("oem name: %s\n", table.oem_name);
  
  printf("sectors per fat: %i\n", table.fat_size);
  printf("sectors per cluster: %i\n", table.sec_per_clust);
  
  printf("root cluster: %i\n", table.root_cluster);
  printf("fat1 start sector: %i\n", table.reserved_sectors);
  
  fcluster = get_cluster32("NTX     IMG");
  
  load_cluster_chain32(fcluster, 0x400000);

}

int fill_fat_info(fat_info_t *fatinfo, char *superblock)
{
  memcpy(fatinfo->oem_name, &superblock[0x03], 8);
  fatinfo->oem_name[8] = 0;
  fatinfo->bytes_per_sec = * (short *) &superblock[0x0B];
  fatinfo->sec_per_clust = superblock[0x0D];
  fatinfo->reserved_sectors = * (short *) &superblock[0x0E];
  fatinfo->numfats = superblock[0x10];
  fatinfo->root_entries = * (short *) &superblock[0x11];
  fatinfo->total_sectors = * (short *) &superblock[0x13];
  fatinfo->media = superblock[0x15];
  
  fatinfo->sec_per_track = * (short *) &superblock[0x18];
  fatinfo->num_heads = * (short *) &superblock[0x1A];
  fatinfo->hidden_sectors = * (int *) &superblock[0x1C];
  fatinfo->total_sectors32 = * (int *) &superblock[0x20];
  
  //fat32 specific
  fatinfo->fat_size = * (int *) &superblock[0x24];
  fatinfo->root_cluster = * (int *) &superblock[0x2C];
}



#define ATA0 0x1F0
#define ATA1 0x170

#define ATAREG_DR 0
#define ATAREG_ER 1
#define ATAREG_FR 1
#define ATAREG_SC 2
#define ATAREG_SN 3
#define ATAREG_CL 4
#define ATAREG_CH 5
#define ATAREG_DH 6
#define ATAREG_DEV 6
#define ATAREG_SR 7
#define ATAREG_CR 7

#define SR_BSY 0x80
#define SR_DRDY 0x40
#define SR_DF 0x20
#define SR_SERV 0x10
#define SR_DRQ 0x8
#define SR_CHK 0x1
#define SR_ERR 0x1




int get_channel (int atanum)
{
  if(atanum == 0) {
    return ATA0;
  } else if (atanum == 1){
    return ATA1;
  }
}

int get_device (int devnum)
{
  if(devnum == 0) {
    return 0xE0;
  } else if(devnum == 1) {
    return 0xF0;
  }
}

int g;

int read_sector (int atanum, int devnum, unsigned int lba, void *buffer)
{
  unsigned int channel;
  unsigned char device;
  unsigned char sr;
  

  channel = get_channel(atanum);
  device = get_device(devnum);

  /*  
  printf("Reading sector %i from %X/%x to %X\n", lba, channel, device);
  g++;
  if(g%2 == 0) {
     tty_cursor = 0;
     g = 0;
    
     }
  */  

  sr = inportb(channel | ATAREG_SR);
  
  while( sr & SR_BSY ||
	 sr & SR_DRQ);
  

  outportb(channel | ATAREG_DEV, device | (lba & 0x0F000000) >> 24);

  sr = inportb(channel | ATAREG_SR);
  
  while( sr & SR_BSY ||
	 sr & SR_DRQ);

  outportb(channel | ATAREG_SC, 1);
  outportb(channel | ATAREG_SN, (lba & 0x000000FF));
  outportb(channel | ATAREG_CL, (lba & 0x0000FF00) >> 8);
  outportb(channel | ATAREG_CH, (lba & 0x00FF0000) >> 16);
  outportb(channel | ATAREG_CR, 0x20);

  while( (inportb(channel | ATAREG_SR) & 8) == 0){
    //    printf("Waiting....\n");
  }

  asm("rep insw"::"D"(buffer),"c"(512/2),"d"(channel | ATAREG_DR));
  
  //  printf("reading ok!\n");
  
}


/*
  NOTE:
  this func assumes FAT12 BPB as defined in bootsect.asm
  (it DOESN'T check it)
*/

char *fat;
char *rootdir;
char *dataarea;

void load_kernel_fat12()
{
  int kclust; //kernel cluster
  
  fat = RAMDISK_START + 512;
  rootdir = fat + 9*2*512;
  dataarea = rootdir + 0xE0*32;
  
  printf("FAT: %X, Root: %X, Data: %X\n",  fat, rootdir, dataarea);
  
  kclust = get_cluster ("NTX     ELF");

  printf("Kernel Cluster: %X\n", kclust);
  
  load_cluster_chain (kclust, 0x110000);

  load_elf(0x110000);
  
  
}

int get_cluster (char *filename)
{
  char name[12];
  char *dir;
  int clust;
  

  dir = rootdir;
  
  for(;;) {
    memcpy(name, dir, 11);
    dir[11] = 0;
    
    if(name[0] != 0) {
      printf("Found name: %s\n", name);
    }
    if(!strcmp(name, filename)){
      break;
    }
    
    dir += 32;
  
    if(dir >= dataarea) {
      printf("File has not been found!!!\n");
      for(;;);
    }
  }
  
  clust = * (short *) (dir + 0x1A);

  return clust;

}

void load_cluster_chain (int cluster, void *dest)
{
  unsigned short wren;
  
  wren = * (unsigned short *) (fat + (cluster*3)/2);
  if( (cluster & 1) == 0) {
    wren &= 0x0FFF;
  } else {
    wren >>= 4;
  }
  
  copy_cluster (cluster, dest);

  if(wren >= 0xFF8) {
    return;
  }
    
  load_cluster_chain(wren, dest + 512);
}

void copy_cluster (int cluster, void *dest)
{
  char *clad; //cluster address

  clad = dataarea + (cluster - 2)*512;
  
  memcpy (dest, clad, 512);
}

void dump(char *addr, int num)
{
  int i;
  puthexi(addr);
  puts(": ");
  for(i=1;i<=num;i++){
    puthex(*(addr+i-1));   
    putchar(' ');
    if(i%16 == 0){
      putchar('\n');
    
      if(i!=num){
	puthexi(addr+i);
	puts(": ");
      }    

    }
  }
}


void *memcpy (void *dest, const void *src, unsigned int n)
{
  int i;
  
  for(i = 0; i < n; i++) {
    *( (char *) dest + i) = *( (char *) src + i);
  }

  return dest;
}
int strcmp(const char *s1, const char *s2)
{
  while (*s1 != 0) {
    if(*s1 == *s2) {
      
    } else {
      
      return *s1 > *s2? 1 : -1;
    
    }
    s1++;
    s2++;
  }
  
  if(*s2 == 0) {
    return 0;
  } else {
    return -1;
  }

}

/* mini tty */


void init_tty()
{
  tty_cursor = 0;
  tty_attribute = 7;
}

int putchar (char c)
{
  char *video = VIDEO_RAM;
  int i;
  int oldx, oldy;

  switch (c) {
  case '\n': //    
    tty_cursor+=VIDEO_WIDTH;
    tty_cursor-=tty_cursor%VIDEO_WIDTH;
    break;
    
  case 8: //backspace
    tty_cursor--;
    *(video + tty_cursor*2) = ' ';
    break;

  default:
    *(video + tty_cursor*2) = c;
    *(video + tty_cursor*2 + 1) = tty_attribute;
    tty_cursor++;
    break;
  }

  //     ,      
  if(tty_cursor>VIDEO_WIDTH*VIDEO_HEIGHT){
    for(i=VIDEO_WIDTH*2;i<=VIDEO_WIDTH*VIDEO_HEIGHT*2+VIDEO_WIDTH*2;i++){
      *(video+i-VIDEO_WIDTH*2)=*(video+i);
    }
    tty_cursor-=VIDEO_WIDTH;
  }

  update_cursor (tty_cursor);
  
}

void update_cursor(unsigned short position)
{
  
  // cursor LOW port to vga INDEX register
  outportb(0x3D4, 0x0F);
  outportb(0x3D5, (unsigned char)(position&0xFF));
  // cursor HIGH port to vga INDEX register
  outportb(0x3D4, 0x0E);                       
  outportb(0x3D5, (unsigned char)((position>>8)&0xFF));
}

inline void outportb(unsigned short port, unsigned char value)
{
  asm("outb %b0,%w1":: "a"(value), "d"(port));
}

inline char inportb(unsigned short port)
{
  char value;

  asm("inb %w1, %b0": "=a"(value): "d"(port));

  return value;
}


void clear()
{
  char *video = (char *) VIDEO_RAM;
  int i;

  for (i = 0; i < VIDEO_HEIGHT*VIDEO_WIDTH; i++) {
    *(video + i*2) = ' ';
    *(video + i*2 + 1) = tty_attribute;
  }
  
  tty_cursor = 0;
  update_cursor(tty_cursor);
}

int textcolor(char attrib)
{
  tty_attribute = attrib;
}

void putdec(unsigned int byte)
{
  unsigned char b1;
  int b[30];
  signed int nb;
  int i=0;
  
  while(1){
    b1=byte%10;
    b[i]=b1;
    nb=byte/10;
    if(nb<=0){
      break;
    }
    i++;
    byte=nb;
  }
  
  for(nb=i+1;nb>0;nb--){
    puthexd(b[nb-1]);
  }
}

void puthexi(unsigned int dword)
{
  puthex( (dword & 0xFF000000) >>24);
  puthex( (dword & 0x00FF0000) >>16);
  puthex( (dword & 0x0000FF00) >>8);
  puthex( (dword & 0x000000FF));
}


void puthex(unsigned char byte)
{
 unsigned  char lb, rb;

  lb=byte >> 4;
  
  rb=byte & 0x0F;
  
  
  puthexd(lb);
  puthexd(rb);
}

void puthexd(unsigned char digit)
{
  char table[]="0123456789ABCDEF";
   putchar(table[digit]);
}


void puts(const char *s)
{
  while(*s) {
    putchar(*s);
    s++;
  }
}
void printf(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);

  dvprintf(fmt, args);

  va_end(args);
}
void dvprintf(const char *fmt, va_list args)
{
  while (*fmt) {
    
    switch (*fmt) {
    case '%':
      fmt++;

      switch (*fmt) {
      case 's':
	puts(va_arg(args, char *));
	break;
	
      case 'c':
	putchar(va_arg(args, unsigned int));
	break;
	
      case 'i':
	putdec(va_arg(args, unsigned int));
	break;

      case 'x':
	puthex(va_arg(args, unsigned int));
	break;
	
      case 'X':
	puthexi(va_arg(args, unsigned int));
	break;
	
      case 'z':
	textcolor(va_arg(args,unsigned int));
	break;

      }
      

      break;
      
    default:
      putchar(*fmt);
      break;
    }

    fmt++;
  }
}
