/*   
 *
 *   Copyright (c) 2003 Lonesome, OSDev Group & Peter Zotov
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */

#include <tyros.h>
#include <scancode.h>
#include <assert.h>
#include <ktty.h>

typedef struct {
  int escaped;
  int shift;
  int alt;
  int ctrl;
} keyboard_state_t;

typedef struct {
  int pointer;
  char buff[0x100];
} tty_buffer_t;


keyboard_state_t *keystate;
tty_buffer_t *tty_buffer;

void ungetc(int c);

void init_keyboard()
{
  unsigned char ks;

  tty_buffer = alloc_pmem(1);
  keystate = alloc_pmem(1);
  
  tty_buffer->pointer = 0;
  keystate->escaped = 0;

  keystate->shift = 0;
  keystate->alt = 0;
  keystate->ctrl = 0;

  printf("Keyboard State: %x\n", inportb(0x64));
  
  while(inportb(0x64) & 1) {
    inportb(0x60);
  }

  outportb(0x21, 0);
  outportb(0x60, 0xF4);
  
}

void receive_scancode (unsigned char scancode)
{
  int ascii;
  int auxsc;

  //  printf("got scancode: %x\n", scancode);

  switch(scancode) {
    
  case 0x36: 
  case 0x2A:
    keystate->shift = 1;
    break;

  case 0x36 + 0x80:
  case 0x2A + 0x80:
    keystate->shift = 0;
    break;

  case 0x1D:
    keystate->ctrl = 1;
    break;

  case 0x1D + 0x80:
    keystate->ctrl = 0;
    break;
    
  default:
    
    if(scancode == 0x46 && keystate->ctrl) {
      //reset terminal
      init_tty();
      clear();
    }

    if(keystate->escaped != 0) {
      
      switch(keystate->escaped) {
	
      case 0xE0:
	
	switch(scancode) {
	case 0x48: //up arrow
	  ungetc(0x1C);
	  break;
	  
	case 0x4B: //left arrow
	  ungetc(0x1D);
	  break;
	  
	case 0x4D: //right arrow
	  ungetc(0x1E);
	  break;
	  
	case 0x50: //down arrow
	  ungetc(0x1F);
	  break;
	}
	break;

      default:

	break;
      }

      keystate->escaped = 0;
    } else {

    if(scancode == 0xE0) {
      keystate->escaped = 0xE0;
    } else {
      
      if(scancode >= 0x80) {
	//	printf("skipping %x\n", scancode);
      } else {
	
	if(keystate->shift){
	  ascii = scancodes_shifted[scancode];
	} else {
	  ascii = scancodes[scancode];
	}
	
	if(ascii == 'c' && keystate->ctrl == 1) {
	  //Ctrl-C
	  action_on_ctrlc(0);
	  break;
	}
	
	
	if(keystate->ctrl) {
	  switch(ascii) {
	  case 'm':
	    ascii = 0x0D;
	  }
	}
	
	
	if(ascii != 0) {
	  ungetc(ascii);
	}
      }
      break;
    }
    
  }

  }

  
}

char getc(void)
{
  char symbol;
  int i;
  task_t *task;

  
    if(tty_buffer->pointer == 0) {
      task = tasks.task[tasks.current];

      task->state = TS_SUSPENDED;
      task->reason_why = RS_WAITING_WHILE_ZERO;
      task->recipient = &tty_buffer->pointer;

      give_up();
      
      assert(task->state != TS_SUSPENDED);
    }
    
  

  isem_disable();

  
  symbol = tty_buffer->buff[0];
  
  for (i = 0; i < tty_buffer->pointer; i++) {
    tty_buffer->buff[i] = tty_buffer->buff[i + 1];
  }

  tty_buffer->pointer--;

  isem_enable();
  

  return symbol;
}


extern int tty_mode;
extern int tty_echo;
void getsn(char *s, int num)
{
  char c = 0;
  char *start;
  
  start = s;

  while ( (c = getc()) != '\n') {
  
    switch(c) {
    case 8:
      if(tty_mode == TTY_MODE_UNBUFFERED) {
	*s = 0x08;
	*(s + 1) = 0;
	return;
	break;
      }
      
      if( start < s) {
	s--;
	*s = 0;
	if(tty_echo) {
	  putchar(c);
	}
      }
      break;
      
    default:

      if ( (s - start) <= num ) {

	*s = c;
	if(tty_echo) {
	  putchar(c);
	}
	s++;
	*s = 0;
      }
      break;
    }
    
    if(tty_mode == TTY_MODE_UNBUFFERED) {
      break;
    }
    
  }
  
  *s = '\n';
  *(s + 1) = 0;
  if(tty_echo) {
    putchar('\n');
  }
  
}

void ungetc(int c)
{
  int handled;

  isem_disable();
  
  handled = -1;
  
  if(hal.mode != 0) {
    handled = gui_receive_keypress(c);
  }
  
  if(handled == -1) {

    tty_buffer->buff[tty_buffer->pointer] = c;
    tty_buffer->pointer++;

  }
  isem_enable();
}
