/*
 *   ktty.c - HAL-level kernel teletype device implementation 
 *
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */  

#include <stdarg.h>

#include <tyros.h>
#include <ktty.h>

void update_cursor(unsigned short position);
void dvprintf(const char *fmt, va_list args);

struct {
  int pointer;
  char buff[0x100];
} tty_buffer;

int tty_attribute;
int tty_cursor;
int tty_escape;
int tty_mode;
int tty_echo;

void init_tty()
{
  tty_attribute = 7;
  tty_cursor = 0;

  tty_buffer.pointer = 0;
  tty_mode = TTY_MODE_BUFFERED;
  tty_echo = 1;
}

void textcolor(char c)
{
  tty_attribute = c;
}

void move(int x, int y)
{
  tty_cursor = y*VIDEO_WIDTH + x;
}

void mvprintf(int x, int y, char *fmt, ...)
{
  int cursor_backup;
  va_list vl;

  cursor_backup = tty_cursor;
  
  move(x,y);
  
  va_start(vl, fmt);

  vprintf(fmt, vl);

  va_end(vl);

  tty_cursor = cursor_backup;
}

void clear()
{
  char *video = (char *) VIDEO_RAM;
  int i;

  for (i = 0; i < VIDEO_HEIGHT*VIDEO_WIDTH; i++) {
    *(video + i*2) = ' ';
    *(video + i*2 + 1) = tty_attribute;
  }
  
  tty_cursor = 0;
  update_cursor(tty_cursor);
}

void mvaddch(int x, int y, char c)
{
  char *video = (char *) VIDEO_RAM;
  *(video + (VIDEO_WIDTH*y + x)*2) = c;
}

void putchar(char c)
{
  hal.putchar(c);
}

void tputchar(char c)
{
  char *video = VIDEO_RAM;
  int i;
  int oldx, oldy;

  if(tty_escape == 0) {
 
  switch (c) {
  case '\n': //    
    tty_cursor+=VIDEO_WIDTH;
    tty_cursor-=tty_cursor%VIDEO_WIDTH;
    break;
    
  case 8: //backspace
    tty_cursor--;
    *(video + tty_cursor*2) = ' ';
    break;

  case 0x0C: //clear
    clear();
    break;

  case 0x0E:
    tty_escape = TTY_ESCAPE_MOVEX;
    break;

  case 0x0F:
    tty_escape = TTY_ESCAPE_MOVEY;
    break;

  case 0x10: //tty buffering ON
    tty_mode = TTY_MODE_BUFFERED;
    break;
  case 0x11: //tty buffering OFF
    tty_mode = TTY_MODE_UNBUFFERED;
    break;
  
  case 0x12: //tty echo ON
    tty_echo = 1;
    break;

  case 0x13: //tty echo OFF
    tty_echo = 0;
    break;

  default:
    *(video + tty_cursor*2) = c;
    *(video + tty_cursor*2 + 1) = tty_attribute;
    tty_cursor++;
    break;
  }

  //     ,      
  if(tty_cursor>VIDEO_WIDTH*VIDEO_HEIGHT){
    for(i=VIDEO_WIDTH*2;i<=VIDEO_WIDTH*VIDEO_HEIGHT*2+VIDEO_WIDTH*2;i++){
      *(video+i-VIDEO_WIDTH*2)=*(video+i);
    }
    tty_cursor-=VIDEO_WIDTH;
  }

  } else {
    switch(tty_escape) {
    case TTY_ESCAPE_MOVEX:
      oldy = tty_cursor/VIDEO_WIDTH;
      tty_cursor = oldy*VIDEO_WIDTH + c;
      break;
      
    case TTY_ESCAPE_MOVEY:
      oldx = tty_cursor%VIDEO_WIDTH;
      tty_cursor = c*VIDEO_WIDTH + oldx;
      break;
    }

    tty_escape = 0;
  }

  update_cursor (tty_cursor);
  
}

void update_cursor(unsigned short position)
{
  
  // cursor LOW port to vga INDEX register
  outportb(0x3D4, 0x0F);
  outportb(0x3D5, (unsigned char)(position&0xFF));
  // cursor HIGH port to vga INDEX register
  outportb(0x3D4, 0x0E);                       
  outportb(0x3D5, (unsigned char)((position>>8)&0xFF));
}



void putdec(unsigned int byte)
{
  unsigned char b1;
  int b[30];
  signed int nb;
  int i=0;
  
  while(1){
    b1=byte%10;
    b[i]=b1;
    nb=byte/10;
    if(nb<=0){
      break;
    }
    i++;
    byte=nb;
  }
  
  for(nb=i+1;nb>0;nb--){
    puthexd(b[nb-1]);
  }
}

void puthexi(unsigned int dword)
{
  puthex( (dword & 0xFF000000) >>24);
  puthex( (dword & 0x00FF0000) >>16);
  puthex( (dword & 0x0000FF00) >>8);
  puthex( (dword & 0x000000FF));
}


void puthex(unsigned char byte)
{
 unsigned  char lb, rb;

  lb=byte >> 4;
  
  rb=byte & 0x0F;
  
  
  puthexd(lb);
  puthexd(rb);
}

void puthexd(unsigned char digit)
{
  char table[]="0123456789ABCDEF";
   putchar(table[digit]);
}


void puts(const char *s)
{
  while(*s) {
    putchar(*s);
    s++;
  }
}
void printf(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);

  dvprintf(fmt, args);

  va_end(args);
}

void dvprintf(const char *fmt, va_list args)
{
  while (*fmt) {
    
    switch (*fmt) {
    case '%':
      fmt++;

      switch (*fmt) {
      case 's':
	puts(va_arg(args, char *));
	break;
	
      case 'c':
	putchar(va_arg(args, unsigned int));
	break;
	
      case 'i':
      case 'd':
	putdec(va_arg(args, unsigned int));
	break;

      case 'x':
	puthex(va_arg(args, unsigned int));
	break;
	
      case 'X':
	puthexi(va_arg(args, unsigned int));
	break;
	
      case 'z':
	textcolor(va_arg(args,unsigned int));
	break;

      }
      

      break;
      
    default:
      putchar(*fmt);
      break;
    }

    fmt++;
  }
}
