/*
 *   mm.c - HAL-level physical memory manager 
 *
 *
 *   Copyright (c) 2003 Lonesome & OSDev Group
 *
 *   http://tyros.sourceforge.net/
 *   http://sourceforge.net/projects/tyros/
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program (see the file COPYING); if not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA
 *
 *
 */  

#include <tyros.h>

#include <assert.h>

#define PHYS_REQUEST_MAX 0x100

//Address the pages bitmap will reside at
#define PAGE_BITMAP 0x300000 

//First 4 megabytes + 1.44 for ramdisk are reserved by the kernel
#define RESERVED_PAGES (1024 + 360)

#define PAGE_ALLOCATED 0x80000000
#define PAGE_FREE 0x00000000

unsigned int page_stack_pointer;
unsigned int *pages_bitmap = (unsigned int *) PAGE_BITMAP;
unsigned int pages;
unsigned int vend_of_heap;

int free_memory;

void *sbrk (int n);


void init_mm()
{
  int i;
  unsigned int free_pages;

  page_stack_pointer = 0;

  vend_of_heap = 0x40000000;

  hal.memory_amount = get_memory_amount();

  pages = hal.memory_amount/0x1000;
  
  
  //printf("System memory amount is %i bytes (%i pages)\n", hal.memory_amount, pages);
  
  free_pages = 0;
  
  for (i = 0; i < pages; i++) {
    
    //if page is reserved then mark it as allocated
    if(i < RESERVED_PAGES) {
      pages_bitmap[i] = PAGE_ALLOCATED;
    } else {
      pages_bitmap[i] = PAGE_FREE;
      free_pages++;
    }
    
  }
  
  free_memory = free_pages*PAGE_SIZE;

  //  printf("Free pages: %i\n", free_pages);

}

#define STEP 0x100000
unsigned int get_memory_amount()
{
  char *p;
  p = (char *) 0x400000;
  do {
    p += STEP;
    *p = 0x77;
  } while (*p == 0x77);
 
  return (unsigned int)  p - STEP;;
}


void free_pmem(unsigned int address)
{
  int page, count, i;

  page = address/0x1000;

  if(page >= pages) {
    printf("free_pmem failure: page >= page\n");
    for(;;);
  }

  if(address & 0xFFF) {
    printf("free_pmem failure: address not aligned\n");
    for(;;);
  }

  count = pages_bitmap[page] & 0x7fffffff;
  
  if(count >= PHYS_REQUEST_MAX) {
    printf("free_pmem failure: count >= PHYS_REQUEST_MAX\n");
    for(;;);
  }

  //  printf("Freeing %X\n", address);

  for(i = 0; i < count; i++) {
    pages_bitmap[page + i] = PAGE_FREE;
  }

  free_memory += count*PAGE_SIZE;
}


int _brk (void *end)
{
  return brk(end);
}

int brk (void *end)
{
  int i;
  unsigned int newal;

  newal = vend_of_heap;
  
  //  printf("BRK called: CVEOH: %X, END: %X\n", newal, end);

  if(end < newal) {
    return -1;
  } else {
    vend_of_heap = end;
    
    for(i = (newal & 0xFFFFF000)  + 0x1000; i <= end; i+=0x1000) {
      //  printf("brk mapping: %X\n", i);
      set_pte (i, hal.pagedir, 0x7, VM_PAGE_ALLOCATED);
    }
    return 0;
  }
}

void *_sbrk(int n)
{
  return sbrk(n);
}



void *sbrk (int n)
{
  unsigned int newal;

  newal = vend_of_heap;


  vend_of_heap += n;
  
  if( (n & 7) + (vend_of_heap & 7) >= 0x1000 ) {
    set_pte (vend_of_heap, hal.pagedir, 0x7, VM_PAGE_ALLOCATED);
    return newal;
  } else { 
    return newal;
  }
}


void *alloc_pmem(int pages_count)
{
  int i, j;
  int freearea;
  int fpage;

  if(pages_count > PHYS_REQUEST_MAX) {
    failure("alloc_pmem: requested more than PHYS_REQUEST_MAX pages\n");
  }
  
  
  freearea = 0;

  for(i = RESERVED_PAGES; i < pages; i++) {
    if (pages_bitmap[i] & PAGE_ALLOCATED) {
      freearea = 0;

    } else {
      freearea++;
      
      if(freearea == pages_count) {
      
	fpage = i - freearea + 1;
	
	

	pages_bitmap[fpage] = PAGE_ALLOCATED | pages_count;
	
	for(j = 1; j < pages_count; j++) {
	  pages_bitmap[fpage + j] = PAGE_ALLOCATED;
	}

	free_memory -= pages_count*PAGE_SIZE;
	
	//	printf("Alloc_Pmem request for %i pages by %i: %X\n", pages_count, tasks.current, fpage *0x1000);
	/*	if((fpage * 0x1000) == 0x4ab000) {
		printf("alloc_pmem breakpoint\n");
		for(;;);
		}*/
	
	return (fpage * 0x1000);
	
      } else {
	//continue
      }
    }
  }
  
  i_disable();
  printf("%zSYSTEM LOW ON MEMORY (REQUESTED %i PAGES)\n", COLOR_LRED, pages_count);
  for(;;);
}
